"""Filter expressions for the vector index request.

The filter expressions are used to filter the results of a vector index request.
This module contains the base class for all filter expressions, as well as
implementations for all the different types of filter expressions::
    And(Equals("foo", "bar"), GreaterThan("age", 18))
    Or(Equals("foo", "bar"), LessThan("age", 18))
    Not(Equals("foo", "bar"))

The `Field` class is used to create filter expressions in a more idiomatic way.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Iterable

from momento_wire_types import vectorindex_pb2 as vectorindex_pb

from momento.errors.exceptions import InvalidArgumentException
from momento.internal.services import Service


@dataclass
class FilterExpression(ABC):
    """Base class for all filter expressions."""

    @abstractmethod
    def to_filter_expression_proto(self) -> vectorindex_pb._FilterExpression:
        """Converts the filter expression to a protobuf filter expression.

        Returns:
            vectorindex_pb._FilterExpression: The protobuf filter expression.
        """
        ...

    def __and__(self, other: FilterExpression) -> FilterExpression:
        """Creates an AND expression between this expression and another.

        Args:
            other (FilterExpression): The other expression to AND with.

        Returns:
            FilterExpression: The AND expression.
        """
        return And(self, other)

    def __or__(self, other: FilterExpression) -> FilterExpression:
        """Creates an OR expression between this expression and another.

        Args:
            other (FilterExpression): The other expression to OR with.

        Returns:
            FilterExpression: The OR expression.
        """
        return Or(self, other)

    def __invert__(self) -> FilterExpression:
        """Creates a NOT expression of this expression.

        Returns:
            FilterExpression: The NOT expression.
        """
        return Not(self)


@dataclass
class And(FilterExpression):
    """Represents an AND expression between two filter expressions."""

    first_expression: FilterExpression
    """The first expression to AND."""
    second_expression: FilterExpression
    """The second expression to AND."""

    def to_filter_expression_proto(self) -> vectorindex_pb._FilterExpression:
        return vectorindex_pb._FilterExpression(and_expression=self.to_proto())

    def to_proto(self) -> vectorindex_pb._AndExpression:
        return vectorindex_pb._AndExpression(
            first_expression=self.first_expression.to_filter_expression_proto(),
            second_expression=self.second_expression.to_filter_expression_proto(),
        )


@dataclass
class Or(FilterExpression):
    """Represents an OR expression between two filter expressions."""

    first_expression: FilterExpression
    """The first expression to OR."""
    second_expression: FilterExpression
    """The second expression to OR."""

    def to_filter_expression_proto(self) -> vectorindex_pb._FilterExpression:
        return vectorindex_pb._FilterExpression(or_expression=self.to_proto())

    def to_proto(self) -> vectorindex_pb._OrExpression:
        return vectorindex_pb._OrExpression(
            first_expression=self.first_expression.to_filter_expression_proto(),
            second_expression=self.second_expression.to_filter_expression_proto(),
        )


@dataclass
class Not(FilterExpression):
    """Represents a NOT expression of a filter expression."""

    expression_to_negate: FilterExpression
    """The expression to negate."""

    def to_filter_expression_proto(self) -> vectorindex_pb._FilterExpression:
        return vectorindex_pb._FilterExpression(not_expression=self.to_proto())

    def to_proto(self) -> vectorindex_pb._NotExpression:
        return vectorindex_pb._NotExpression(
            expression_to_negate=self.expression_to_negate.to_filter_expression_proto()
        )


@dataclass
class Equals(FilterExpression):
    """Represents an equals expression between a field and a value."""

    field: str
    """The field to compare."""
    value: str | int | float | bool
    """The value to test equality with."""

    def to_filter_expression_proto(self) -> vectorindex_pb._FilterExpression:
        return vectorindex_pb._FilterExpression(equals_expression=self.to_proto())

    def to_proto(self) -> vectorindex_pb._EqualsExpression:
        if type(self.value) is str:
            return vectorindex_pb._EqualsExpression(field=self.field, string_value=self.value)
        elif type(self.value) is int:
            return vectorindex_pb._EqualsExpression(field=self.field, integer_value=self.value)
        elif type(self.value) is float:
            return vectorindex_pb._EqualsExpression(field=self.field, float_value=self.value)
        elif type(self.value) is bool:
            return vectorindex_pb._EqualsExpression(field=self.field, boolean_value=self.value)
        else:
            raise InvalidArgumentException(
                f"Invalid type for value: {type(self.value)} in equals expression. Must be one of str, int, float, bool.",
                Service.INDEX,
            )


@dataclass
class GreaterThan(FilterExpression):
    """Represents a greater than expression between a field and a value."""

    field: str
    """The field to compare."""
    value: int | float
    """The value to test greater than with."""

    def to_filter_expression_proto(self) -> vectorindex_pb._FilterExpression:
        return vectorindex_pb._FilterExpression(greater_than_expression=self.to_proto())

    def to_proto(self) -> vectorindex_pb._GreaterThanExpression:
        if type(self.value) is int:
            return vectorindex_pb._GreaterThanExpression(field=self.field, integer_value=self.value)
        elif type(self.value) is float:
            return vectorindex_pb._GreaterThanExpression(field=self.field, float_value=self.value)
        else:
            raise InvalidArgumentException(
                f"Invalid type for value: {type(self.value)} in greater than expression. Must be one of int, float.",
                Service.INDEX,
            )


@dataclass
class GreaterThanOrEqual(FilterExpression):
    """Represents a greater than or equal expression between a field and a value."""

    field: str
    """The field to compare."""
    value: int | float
    """The value to test greater than or equal with."""

    def to_filter_expression_proto(self) -> vectorindex_pb._FilterExpression:
        return vectorindex_pb._FilterExpression(greater_than_or_equal_expression=self.to_proto())

    def to_proto(self) -> vectorindex_pb._GreaterThanOrEqualExpression:
        if type(self.value) is int:
            return vectorindex_pb._GreaterThanOrEqualExpression(field=self.field, integer_value=self.value)
        elif type(self.value) is float:
            return vectorindex_pb._GreaterThanOrEqualExpression(field=self.field, float_value=self.value)
        else:
            raise InvalidArgumentException(
                f"Invalid type for value: {type(self.value)} in greater than or equal expression. Must be one of int, float.",
                Service.INDEX,
            )


@dataclass
class LessThan(FilterExpression):
    """Represents a less than expression between a field and a value."""

    field: str
    """The field to compare."""
    value: int | float
    """The value to test less than with."""

    def to_filter_expression_proto(self) -> vectorindex_pb._FilterExpression:
        return vectorindex_pb._FilterExpression(less_than_expression=self.to_proto())

    def to_proto(self) -> vectorindex_pb._LessThanExpression:
        if type(self.value) is int:
            return vectorindex_pb._LessThanExpression(field=self.field, integer_value=self.value)
        elif type(self.value) is float:
            return vectorindex_pb._LessThanExpression(field=self.field, float_value=self.value)
        else:
            raise InvalidArgumentException(
                f"Invalid type for value: {type(self.value)} in less than expression. Must be one of int, float.",
                Service.INDEX,
            )


@dataclass
class LessThanOrEqual(FilterExpression):
    """Represents a less than or equal expression between a field and a value."""

    field: str
    """The field to compare."""
    value: int | float
    """The value to test less than or equal with."""

    def to_filter_expression_proto(self) -> vectorindex_pb._FilterExpression:
        return vectorindex_pb._FilterExpression(less_than_or_equal_expression=self.to_proto())

    def to_proto(self) -> vectorindex_pb._LessThanOrEqualExpression:
        if type(self.value) is int:
            return vectorindex_pb._LessThanOrEqualExpression(field=self.field, integer_value=self.value)
        elif type(self.value) is float:
            return vectorindex_pb._LessThanOrEqualExpression(field=self.field, float_value=self.value)
        else:
            raise InvalidArgumentException(
                f"Invalid type for value: {type(self.value)} in less than or equal expression. Must be one of int, float.",
                Service.INDEX,
            )


@dataclass
class ListContains(FilterExpression):
    """Represents a list contains expression between a field and a value."""

    field: str
    """The list field to test."""
    value: str
    """The value to test list contains with."""

    def to_filter_expression_proto(self) -> vectorindex_pb._FilterExpression:
        return vectorindex_pb._FilterExpression(list_contains_expression=self.to_proto())

    def to_proto(self) -> vectorindex_pb._ListContainsExpression:
        # todo should make oneof defensively
        return vectorindex_pb._ListContainsExpression(field=self.field, string_value=self.value)


@dataclass
class IdInSet(FilterExpression):
    """Represents an expression to test if an item id is in a set of ids."""

    ids: Iterable[str]
    """The set of ids to test id in set with."""

    def to_filter_expression_proto(self) -> vectorindex_pb._FilterExpression:
        return vectorindex_pb._FilterExpression(id_in_set_expression=self.to_proto())

    def to_proto(self) -> vectorindex_pb._IdInSetExpression:
        return vectorindex_pb._IdInSetExpression(ids=list(self.ids))



from __future__ import annotations

import math

from momento_wire_types import cachepubsub_pb2 as pubsub_pb
from momento_wire_types import cachepubsub_pb2_grpc as pubsub_grpc

from momento import logs
from momento.auth import CredentialProvider
from momento.config import TopicConfiguration
from momento.errors import convert_error
from momento.internal._utilities import _validate_cache_name, _validate_topic_name
from momento.internal.services import Service
from momento.internal.synchronous._scs_grpc_manager import (
    _PubsubGrpcManager,
    _PubsubGrpcStreamManager,
)
from momento.responses import (
    TopicPublish,
    TopicPublishResponse,
    TopicSubscribe,
    TopicSubscribeResponse,
)


class _ScsPubsubClient:
    """Internal pubsub client."""

    stream_topic_manager_count = 0

    def __init__(self, configuration: TopicConfiguration, credential_provider: CredentialProvider):
        endpoint = credential_provider.cache_endpoint
        self._logger = logs.logger
        self._logger.debug("Pubsub client instantiated with endpoint: %s", endpoint)
        self._endpoint = endpoint

        num_subscriptions = configuration.get_max_subscriptions()
        # Default to a single channel and scale up if necessary. Each channel can support
        # 100 subscriptions. Issuing more subscribe requests than you have channels to handle
        # will cause the last request to hang indefinitely, so it's important to get this right.
        num_channels = 1
        if num_subscriptions > 0:
            num_channels = math.ceil(num_subscriptions / 100.0)
            self._logger.debug(f"creating {num_channels} subscription channels")

        self._grpc_manager = _PubsubGrpcManager(configuration, credential_provider)
        self._stream_managers = [
            _PubsubGrpcStreamManager(configuration, credential_provider) for i in range(0, num_channels)
        ]

    @property
    def endpoint(self) -> str:
        return self._endpoint

    def publish(self, cache_name: str, topic_name: str, value: str | bytes) -> TopicPublishResponse:
        try:
            _validate_cache_name(cache_name)
            _validate_topic_name(topic_name)

            if isinstance(value, str):
                topic_value = pubsub_pb._TopicValue(text=value)
            else:
                topic_value = pubsub_pb._TopicValue(binary=value)

            request = pubsub_pb._PublishRequest(
                cache_name=cache_name,
                topic=topic_name,
                value=topic_value,
            )

            self._get_stub().Publish(
                request,
            )
            return TopicPublish.Success()
        except Exception as e:
            self._log_request_error("publish", e)
            return TopicPublish.Error(convert_error(e, Service.TOPICS))

    def subscribe(self, cache_name: str, topic_name: str) -> TopicSubscribeResponse:
        try:
            _validate_cache_name(cache_name)
            _validate_topic_name(topic_name)

            request = pubsub_pb._SubscriptionRequest(
                cache_name=cache_name,
                topic=topic_name,
                # TODO: resume_at_topic_sequence_number
            )
            stream = self._get_stream_stub().Subscribe(  # type: ignore[misc]
                request,
            )

            # Ping the stream to provide a nice error message if the cache does not exist.
            msg: pubsub_pb._SubscriptionItem = stream.next()  # type: ignore[misc]
            msg_type: str = msg.WhichOneof("kind")
            if msg_type == "heartbeat":
                # The first message to a new subscription is always a heartbeat.
                pass
            else:
                err = Exception(f"expected a heartbeat message but got '{msg_type}'")
                self._log_request_error("subscribe", err)
                return TopicSubscribe.Error(convert_error(err, Service.TOPICS))
            return TopicSubscribe.Subscription(cache_name, topic_name, client_stream=stream)  # type: ignore[misc]
        except Exception as e:
            self._log_request_error("subscribe", e)
            return TopicSubscribe.Error(convert_error(e, Service.TOPICS))

    def _log_request_error(self, request_type: str, e: Exception) -> None:
        self._logger.warning(f"{request_type} failed with exception: {e}")

    def _get_stub(self) -> pubsub_grpc.PubsubStub:
        return self._grpc_manager.stub()

    def _get_stream_stub(self) -> pubsub_grpc.PubsubStub:
        stub = self._stream_managers[self.stream_topic_manager_count % len(self._stream_managers)].stub()
        self.stream_topic_manager_count += 1
        return stub

    def close(self) -> None:
        self._grpc_manager.close()
        for stream_client in self._stream_managers:
            stream_client.close()



from __future__ import annotations

from threading import Event
from typing import Optional

import grpc
from momento_wire_types import cacheclient_pb2_grpc as cache_client
from momento_wire_types import cachepubsub_pb2_grpc as pubsub_client
from momento_wire_types import controlclient_pb2_grpc as control_client

from momento import logs
from momento.auth import CredentialProvider
from momento.config import Configuration, TopicConfiguration
from momento.internal._utilities import momento_version
from momento.internal._utilities._channel_credentials import (
    channel_credentials_from_root_certs_or_default,
)
from momento.internal.synchronous._add_header_client_interceptor import (
    AddHeaderClientInterceptor,
    AddHeaderStreamingClientInterceptor,
    Header,
)
from momento.internal.synchronous._retry_interceptor import RetryInterceptor
from momento.retry import RetryStrategy


class _ControlGrpcManager:
    """Internal gRPC control mananger."""

    version = momento_version

    def __init__(self, configuration: Configuration, credential_provider: CredentialProvider):
        self._secure_channel = grpc.secure_channel(
            target=credential_provider.control_endpoint,
            credentials=channel_credentials_from_root_certs_or_default(configuration),
        )
        intercept_channel = grpc.intercept_channel(
            self._secure_channel, *_interceptors(credential_provider.auth_token, configuration.get_retry_strategy())
        )
        self._stub = control_client.ScsControlStub(intercept_channel)  # type: ignore[no-untyped-call]

    def close(self) -> None:
        self._secure_channel.close()

    def stub(self) -> control_client.ScsControlStub:
        return self._stub


class _DataGrpcManager:
    """Internal gRPC data mananger."""

    version = momento_version

    def __init__(self, configuration: Configuration, credential_provider: CredentialProvider):
        self._logger = logs.logger
        self._secure_channel = grpc.secure_channel(
            target=credential_provider.cache_endpoint,
            credentials=channel_credentials_from_root_certs_or_default(configuration),
        )

        intercept_channel = grpc.intercept_channel(
            self._secure_channel, *_interceptors(credential_provider.auth_token, configuration.get_retry_strategy())
        )
        self._stub = cache_client.ScsStub(intercept_channel)  # type: ignore[no-untyped-call]

    """
        This method tries to eagerly connect to Momento's server until
        EAGER_CONNECTION_TIMEOUT elapses.
    """

    def eagerly_connect(self, timeout_seconds: float) -> None:
        self._logger.debug(
            "Attempting to create an eager connection with Momento's server within " f"{timeout_seconds} seconds"
        )

        # An event to track whether we were able to establish an eager connection
        # This is required as we create a subscription to eagerly connect and observe the state changes
        # We do NOT want the subscription to lurk around after it's job is done.
        connection_event = Event()

        def on_timeout() -> None:
            self._logger.debug(
                "We could not establish an eager connection within %d seconds",
                timeout_seconds,
            )
            # the subscription is no longer needed; it was only meant to watch if we could connect eagerly
            self._secure_channel.unsubscribe(on_state_change)

        """
        A callback that is triggered whenever a connection's state changes. We explicitly subscribe to
        to the channel to notify us of state transitions. This method essentially handles unsubscribing
        as soon as we reach the desired state (or an unexpected one). In theory this callback isn't needed
        to eagerly connect, but we still need it to not have a lurking subscription.
        """

        def on_state_change(state: grpc.ChannelConnectivity) -> None:
            ready: grpc.ChannelConnectivity = grpc.ChannelConnectivity.READY
            connecting: grpc.ChannelConnectivity = grpc.ChannelConnectivity.CONNECTING
            idle: grpc.ChannelConnectivity = grpc.ChannelConnectivity.IDLE
            if state == ready:
                self._logger.debug("Connected to Momento's server! Happy Caching!")
                # we successfully connected within the timeout and we no longer need this subscription
                self._secure_channel.unsubscribe(on_state_change)
                # this indicates to the connection event that we were successful in establishing an eager connection
                connection_event.set()

            elif state == idle:
                self._logger.debug("State is idle; waiting to transition to CONNECTING")
            elif state == connecting:
                self._logger.debug("State transitioned to CONNECTING; waiting to get READY")
            else:
                self._logger.warn(f"Unexpected connection state: {state}. while trying to eagerly connect")
                # we could not connect within the timeout and we no longer need this subscription
                self._secure_channel.unsubscribe(on_state_change)
                connection_event.set()

        # we subscribe to the channel that notifies us of state transitions, and the connection event above will take
        # care of unsubscribing from the channel incase the timeout has elapsed.
        self._secure_channel.subscribe(on_state_change, try_to_connect=True)

        connection_established = connection_event.wait(timeout_seconds)
        if not connection_established:
            on_timeout()

    def close(self) -> None:
        self._logger.debug("Closing and tearing down gRPC channel")
        self._secure_channel.close()

    def stub(self) -> cache_client.ScsStub:
        return self._stub


class _PubsubGrpcManager:
    """Internal gRPC pubsub manager."""

    version = momento_version

    def __init__(self, configuration: TopicConfiguration, credential_provider: CredentialProvider):
        self._secure_channel = grpc.secure_channel(
            target=credential_provider.cache_endpoint,
            credentials=grpc.ssl_channel_credentials(),
        )
        intercept_channel = grpc.intercept_channel(
            self._secure_channel, *_interceptors(credential_provider.auth_token, None)
        )
        self._stub = pubsub_client.PubsubStub(intercept_channel)  # type: ignore[no-untyped-call]

    def close(self) -> None:
        self._secure_channel.close()

    def stub(self) -> pubsub_client.PubsubStub:
        return self._stub


class _PubsubGrpcStreamManager:
    """Internal gRPC pubsub stream manager."""

    version = momento_version

    def __init__(self, configuration: TopicConfiguration, credential_provider: CredentialProvider):
        self._secure_channel = grpc.secure_channel(
            target=credential_provider.cache_endpoint,
            credentials=grpc.ssl_channel_credentials(),
        )
        intercept_channel = grpc.intercept_channel(
            self._secure_channel, *_stream_interceptors(credential_provider.auth_token)
        )
        self._stub = pubsub_client.PubsubStub(intercept_channel)  # type: ignore[no-untyped-call]

    def close(self) -> None:
        self._secure_channel.close()

    def stub(self) -> pubsub_client.PubsubStub:
        return self._stub


def _interceptors(
    auth_token: str, retry_strategy: Optional[RetryStrategy] = None
) -> list[grpc.UnaryUnaryClientInterceptor]:
    headers = [Header("authorization", auth_token), Header("agent", f"python:{_ControlGrpcManager.version}")]
    return list(
        filter(
            None, [AddHeaderClientInterceptor(headers), RetryInterceptor(retry_strategy) if retry_strategy else None]
        )
    )


def _stream_interceptors(auth_token: str) -> list[grpc.UnaryStreamClientInterceptor]:
    headers = [
        Header("authorization", auth_token),
        Header("agent", f"python:{_PubsubGrpcStreamManager.version}"),
    ]
    return [AddHeaderStreamingClientInterceptor(headers)]



from momento.common_data.vector_index.item import Item

from .control.create import CreateIndex, CreateIndexResponse
from .control.delete import DeleteIndex, DeleteIndexResponse
from .control.list import IndexInfo, ListIndexes, ListIndexesResponse
from .data.count_items import CountItems, CountItemsResponse
from .data.delete_item_batch import DeleteItemBatch, DeleteItemBatchResponse
from .data.get_item_batch import (
    GetItemBatch,
    GetItemBatchResponse,
)
from .data.get_item_metadata_batch import GetItemMetadataBatch, GetItemMetadataBatchResponse
from .data.search import Search, SearchHit, SearchResponse
from .data.search_and_fetch_vectors import (
    SearchAndFetchVectors,
    SearchAndFetchVectorsHit,
    SearchAndFetchVectorsResponse,
)
from .data.upsert_item_batch import UpsertItemBatch, UpsertItemBatchResponse
from .response import VectorIndexResponse

__all__ = [
    "CountItems",
    "CountItemsResponse",
    "CreateIndex",
    "CreateIndexResponse",
    "DeleteIndex",
    "DeleteIndexResponse",
    "GetItemMetadataBatch",
    "GetItemMetadataBatchResponse",
    "GetItemBatch",
    "GetItemBatchResponse",
    "IndexInfo",
    "Item",
    "ListIndexes",
    "ListIndexesResponse",
    "SearchHit",
    "Search",
    "SearchAndFetchVectors",
    "SearchAndFetchVectorsHit",
    "SearchResponse",
    "SearchAndFetchVectorsResponse",
    "UpsertItemBatch",
    "UpsertItemBatchResponse",
    "DeleteItemBatch",
    "DeleteItemBatchResponse",
    "VectorIndexResponse",
]
